import warnings
from typing import Optional, cast

import torch
from torch.nn.functional import relu

__all__ = [
    "nonnegsqrt",
    "gram_matrix",
    "cosine_similarity",
]


def nonnegsqrt(x: torch.Tensor) -> torch.Tensor:
    r"""Safely calculates the square-root of a non-negative input

    .. math::

        \fun{nonnegsqrt}{x} =
        \begin{cases}
            \sqrt{x} &\quad\text{if } x \ge 0 \\
            0 &\quad\text{otherwise}
        \end{cases}

    .. note::

        This operation is useful in situations where the input tensor is strictly
        non-negative from a theoretical standpoint, but might be negative due to
        numerical instabilities.

    Args:
        x: Input tensor.
    """
    return torch.sqrt(relu(x))


def gram_matrix(x: torch.Tensor, normalize: bool = False) -> torch.Tensor:
    r"""Calculates the channel-wise `Gram matrix
        <https://en.wikipedia.org/wiki/Gramian_matrix>`_ of a batched input tensor.

        Given a tensor :math:`x` of shape
        :math:`B \times C \times N_1 \times \dots \times N_D` each element of the
        single-sample Gram matrix :math:`G_{b,c_1 c_2}` with :math:`b \in 1,\dots,B` and
        :math:`c_1,\,c_2  \in 1,\dots,C` is calculated by

        .. math::

            G_{b,c_1 c_2} = \dotproduct{\fun{vec}{x_{b, c_1}}}{\fun{vec}{x_{b, c_2}}}

        where :math:`\dotproduct{\cdot}{\cdot}` denotes the
        `dot product <https://en.wikipedia.org/wiki/Dot_product>`_ and
        :math:`\fun{vec}{\cdot}` denotes the
        `vectorization function <https://en.wikipedia.org/wiki/Vectorization_(mathematics)>`_ .

        Args:
            x: Input tensor of shape :math:`B \times C \times N_1 \times \dots \times N_D`
            normalize: If True, normalizes the Gram matrix :math:`G` by
                :math:`\prod\limits_{d=1}^{D} N_d` to keep the value range similar for
                different sized inputs. Defaults to ``False``.

        Returns:
            Channel-wise Gram matrix G of shape :math:`B \times C \times C`.
        """
    x_flat = torch.flatten(x, 2)
    gram_matrix = torch.bmm(x_flat, x_flat.transpose(1, 2))
    if not normalize:
        return gram_matrix

    numel = x_flat.size()[-1]
    return gram_matrix / numel


def _norm(x: torch.Tensor, dim: int = 1, eps: float = 1e-8) -> torch.Tensor:
    return cast(torch.Tensor, x / (torch.norm(x, dim=dim, keepdim=True) + eps))


def cosine_similarity(
    x1: torch.Tensor,
    x2: torch.Tensor,
    eps: float = 1e-8,
    batched_input: Optional[bool] = None,
) -> torch.Tensor:
    r"""Calculates the cosine similarity between the samples of ``x1`` and ``x2``.

    Args:
        x1: First input of shape :math:`S_1 \times N_1 \times \dots \times N_D`.
        x2: Second input of shape :math:`S_2 \times N_1 \times \dots \times N_D`.
        eps: Small value to avoid zero division. Defaults to ``1e-8``.
        batched_input: If ``True``, treat the first dimension of the inputs as batch
            dimension, i.e. :math:`B \times S \times N_1 \times \dots \times N_D`.
            Defaults to ``False``.

    Returns:
        Similarity matrix of shape :math:`S_1 \times S_2` in which every element
        represents the cosine similarity between the corresponding samples :math:`S` of
        ``x1`` and ``x2``. If ``batched_input is True``, the output shape is
        :math:`B \times S_1 \times S_2`

    Note:
        The default value of ``batched_input`` will change from ``False`` to ``True``
        in the future.

    """
    if batched_input is None:
        msg = (
            "The default value of batched_input will change "
            "from False to True in the future."
        )
        warnings.warn(msg, FutureWarning)
        batched_input = False

    mm, dim = (torch.bmm, 2) if batched_input else (torch.mm, 1)

    x1 = _norm(torch.flatten(x1, dim), dim=dim, eps=eps)
    x2 = _norm(torch.flatten(x2, dim), dim=dim, eps=eps)
    return torch.clamp(mm(x1, x2.transpose(-1, -2)), max=1.0 / eps)



from typing import Optional, Sequence, Tuple, cast

import torch

from pystiche.image.utils import extract_num_channels, force_batched_image

__all__ = [
    "normalize",
    "denormalize",
    "float_to_uint8_range",
    "uint8_to_float_range",
    "reverse_channel_order",
    "transform_channels_linearly",
    "transform_channels_affinely",
]


def float_to_uint8_range(x: torch.Tensor) -> torch.Tensor:
    return x * 255.0


def uint8_to_float_range(x: torch.Tensor) -> torch.Tensor:
    return x / 255.0


@force_batched_image
def reverse_channel_order(x: torch.Tensor) -> torch.Tensor:
    return x.flip((1,))


def _channel_stats_to_tensor(
    image: torch.Tensor, mean: Sequence[float], std: Sequence[float]
) -> Tuple[torch.Tensor, torch.Tensor]:
    num_channels = extract_num_channels(image)

    def to_tensor(seq: Sequence[float]) -> torch.Tensor:
        if len(seq) != num_channels:
            msg = (
                f"The length of the channel statistics and the number of image "
                f"channels do not match: {len(seq)} != {num_channels}"
            )
            raise RuntimeError(msg)
        return torch.tensor(seq, device=image.device).view(1, -1, 1, 1)

    return to_tensor(mean), to_tensor(std)


@force_batched_image
def normalize(
    x: torch.Tensor, mean: Sequence[float], std: Sequence[float]
) -> torch.Tensor:
    mean, std = _channel_stats_to_tensor(x, mean, std)
    return (x - mean) / std


@force_batched_image
def denormalize(
    x: torch.Tensor, mean: Sequence[float], std: Sequence[float]
) -> torch.Tensor:
    mean, std = _channel_stats_to_tensor(x, mean, std)
    return x * std + mean


@force_batched_image
def transform_channels_affinely(
    x: torch.Tensor, matrix: torch.Tensor, bias: Optional[torch.Tensor] = None
) -> torch.Tensor:
    batch_size, _, *spatial_size = x.size()
    x = torch.flatten(x, 2)

    if matrix.dim() == 2:
        matrix = matrix.unsqueeze(0).repeat(batch_size, 1, 1)
    num_channels = matrix.size()[1]

    x = torch.bmm(matrix, x)

    if bias is not None:
        if bias.dim() == 2:
            bias = bias.unsqueeze(0)
        x += bias

    return x.view(batch_size, num_channels, *spatial_size)


def transform_channels_linearly(x: torch.Tensor, matrix: torch.Tensor) -> torch.Tensor:
    return cast(torch.Tensor, transform_channels_affinely(x, matrix, bias=None))



import time
import warnings
from types import TracebackType
from typing import Any, Callable, Iterable, Iterator, Optional, Tuple, Type, Union, cast

import torch
from torch import nn, optim
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader

import pystiche
from pystiche import loss
from pystiche.image import extract_aspect_ratio, extract_image_size
from pystiche.misc import build_deprecation_message, suppress_warnings
from pystiche.pyramid import ImagePyramid
from pystiche.pyramid.level import PyramidLevel

with suppress_warnings():
    from .log import (
        OptimLogger,
        default_epoch_header,
        default_image_optim_log_fn,
        default_pyramid_level_header,
        default_transformer_optim_log_fn,
    )

__all__ = [
    "default_image_optimizer",
    "image_optimization",
    "pyramid_image_optimization",
    "default_model_optimizer",
    "model_optimization",
    "multi_epoch_model_optimization",
    "default_image_optim_loop",
    "default_image_pyramid_optim_loop",
    "default_transformer_optimizer",
    "default_transformer_optim_loop",
    "default_transformer_epoch_optim_loop",
]


def default_image_optimizer(input_image: torch.Tensor) -> optim.LBFGS:
    r"""
    Args:
        input_image: Image to be optimized.

    Returns:
        :class:`torch.optim.LBFGS` optimizer with a learning rate of ``1.0``. The
        pixels of ``input_image`` are set as optimization parameters.
    """
    return optim.LBFGS([input_image.requires_grad_(True)], lr=1.0, max_iter=1)


# Unfortunately contextlib.nullcontext is only available since Python 3.7
# https://docs.python.org/3.7/library/contextlib.html#contextlib.nullcontext
class _NullContext:
    def __call__(self, input_image: torch.Tensor) -> "_NullContext":
        return self

    def __enter__(self) -> None:
        pass

    def __exit__(
        self, exc_type: Type[Exception], exc_val: Exception, exc_tb: TracebackType,
    ) -> None:
        pass


def _log_parameter_deprecation(name: str) -> None:
    msg = build_deprecation_message(
        f"The parameter {name}",
        "0.7.0",
        info="See https://github.com/pmeier/pystiche/issues/434 for details.",
    )
    warnings.warn(msg, UserWarning)


def image_optimization(
    input_image: torch.Tensor,
    criterion: nn.Module,
    optimizer: Optional[Union[Optimizer, Callable[[torch.Tensor], Optimizer]]] = None,
    get_optimizer: Optional[Callable[[torch.Tensor], Optimizer]] = None,
    num_steps: Union[int, Iterable[int]] = 500,
    preprocessor: Optional[nn.Module] = None,
    postprocessor: Optional[nn.Module] = None,
    quiet: bool = False,
    logger: Optional[OptimLogger] = None,
    log_fn: Optional[
        Callable[[int, Union[torch.Tensor, pystiche.LossDict]], None]
    ] = None,
) -> torch.Tensor:
    r"""Perform an image optimization with integrated logging.

    Args:
        input_image: Image to be optimized.
        criterion: Optimization criterion.
        optimizer: Optional optimizer or optimizer getter. If omitted,
            :func:`default_image_optimizer` is used. If a ``preprocessor`` is used, has
            to be a getter.
        num_steps: Number of optimization steps. Defaults to ``500``.
        preprocessor: Optional preprocessor that is called with the ``input_image``
            before the optimization.
        postprocessor: Optional preprocessor that is called with the ``input_image``
            after the optimization.
        quiet: If ``True``, not information is logged during the optimization. Defaults
            to ``False``.
        logger: Optional custom logger. If ``None``,
            :class:`pystiche.optim.OptimLogger` is used. Defaults to ``None``.
        log_fn: Optional custom logging function. It is called in every optimization
            step with the current step and loss. If ``None``,
            :func:`pystiche.optim.default_image_optim_log_fn` is used. Defaults to
            ``None``.

    Raises:
        RuntimeError: If ``preprocessor`` is used and ``optimizer`` is not passed as
        getter.
    """
    if get_optimizer is not None:
        msg = build_deprecation_message(
            "The parameter get_optimizer", "0.7.0", info="It was renamed to optimizer"
        )
        warnings.warn(msg)
        optimizer = get_optimizer

    if isinstance(optimizer, Optimizer) and preprocessor:
        raise RuntimeError(
            "If a preprocessor is used, optimizer has to be passed as getter"
        )

    if optimizer is None:
        optimizer = default_image_optimizer

    if isinstance(num_steps, int):
        num_steps = range(1, num_steps + 1)

    if logger is None:
        with suppress_warnings():
            logger = OptimLogger()
    else:
        _log_parameter_deprecation("logger")

    if log_fn is None:
        with suppress_warnings():
            log_fn = default_image_optim_log_fn(optim_logger=logger)
    else:
        _log_parameter_deprecation("log_fn")

    if preprocessor:
        with torch.no_grad():
            input_image = preprocessor(input_image)

    if not isinstance(optimizer, Optimizer):
        optimizer = optimizer(input_image)

    mle_handler = (
        loss.MLEHandler(criterion)
        if not isinstance(criterion, loss.MultiOperatorLoss)
        else _NullContext()
    )

    for step in num_steps:

        def closure() -> float:
            # See https://github.com/pmeier/pystiche/pull/264#discussion_r430205029
            optimizer.zero_grad()  # type: ignore[union-attr]

            with mle_handler(input_image):  # type: ignore[operator]
                loss = criterion(input_image)
            loss.backward()

            if not quiet:
                # See https://github.com/pmeier/pystiche/pull/264#discussion_r430205029
                log_fn(step, loss)  # type: ignore[misc]

            return cast(float, loss.item())

        optimizer.step(closure)

    if postprocessor:
        with torch.no_grad():
            input_image = postprocessor(input_image)

    return input_image.detach()


def default_image_optim_loop(*args: Any, **kwargs: Any) -> Any:
    msg = build_deprecation_message(
        "The function default_image_optim_loop",
        "0.7.0",
        info="It was renamed to image_optimization",
    )
    warnings.warn(msg)
    return image_optimization(*args, **kwargs)


def pyramid_image_optimization(
    input_image: torch.Tensor,
    criterion: nn.Module,
    pyramid: ImagePyramid,
    get_optimizer: Optional[Callable[[torch.Tensor], Optimizer]] = None,
    preprocessor: Optional[nn.Module] = None,
    postprocessor: Optional[nn.Module] = None,
    quiet: bool = False,
    logger: Optional[OptimLogger] = None,
    get_pyramid_level_header: Optional[
        Callable[[int, PyramidLevel, Tuple[int, int]], str]
    ] = None,
    log_fn: Optional[
        Callable[[int, Union[torch.Tensor, pystiche.LossDict]], None]
    ] = None,
) -> torch.Tensor:
    r"""Perform a image optimization for :class:`pystiche.pyramid.ImagePyramid` s with
    integrated logging.

    Args:
        input_image: Image to be optimized.
        criterion: Optimization criterion.
        pyramid: Image pyramid.
        get_optimizer: Optional getter for the optimizer. If ``None``,
            :func:`default_image_optimizer` is used. Defaults to ``None``.
        preprocessor: Optional preprocessor that is called with the ``input_image``
            before the optimization.
        postprocessor: Optional preprocessor that is called with the ``input_image``
            after the optimization.
        quiet: If ``True``, not information is logged during the optimization. Defaults
            to ``False``.
        logger: Optional custom logger. If ``None``,
            :class:`pystiche.optim.OptimLogger` is used. Defaults to ``None``.
        get_pyramid_level_header: Optional custom getter for the logged pyramid level
            header. It is called before each level with the current level number,
            the :class:`pystiche.pyramid.PyramidLevel`, and the size of the
            ``input_image``. If ``None``
            :func:`pystiche.optim.default_pyramid_level_header` is used. Defaults to
            ``None``.
        log_fn: Optional custom logging function. It is called in every optimization
            step with the current step and loss. If ``None``,
            :func:`pystiche.optim.default_image_optim_log_fn` is used. Defaults to
            ``None``.
    """
    aspect_ratio = extract_aspect_ratio(input_image)
    if get_optimizer is None:
        get_optimizer = default_image_optimizer

    if logger is None:
        with suppress_warnings():
            logger = OptimLogger()
    else:
        _log_parameter_deprecation("logger")

    if get_pyramid_level_header is None:
        get_pyramid_level_header = default_pyramid_level_header
    else:
        _log_parameter_deprecation("get_pyramid_level_header")

    output_image = input_image
    for num, level in enumerate(pyramid, 1):

        def optimization(input_image: torch.Tensor) -> torch.Tensor:
            return image_optimization(
                input_image,
                criterion,
                optimizer=get_optimizer,
                num_steps=iter(level),
                preprocessor=preprocessor,
                postprocessor=postprocessor,
                quiet=quiet,
                logger=logger,
                log_fn=log_fn,
            )

        with torch.no_grad():
            input_image = level.resize_image(output_image, aspect_ratio=aspect_ratio)

        if quiet:
            output_image = optimization(input_image)
        else:
            input_image_size = extract_image_size(input_image)
            with suppress_warnings():
                header = get_pyramid_level_header(num, level, input_image_size)
            with logger.environment(header):
                output_image = optimization(input_image)

    return output_image


def default_image_pyramid_optim_loop(*args: Any, **kwargs: Any) -> Any:
    msg = build_deprecation_message(
        "The function default_image_pyramid_optim_loop",
        "0.7.0",
        info="It was renamed to pyramid_image_optimization",
    )
    warnings.warn(msg)
    return pyramid_image_optimization(*args, **kwargs)


def default_model_optimizer(transformer: nn.Module) -> Optimizer:
    r"""
    Args:
        transformer: Transformer to be optimized.

    Returns:
        :class:`torch.optim.Adam` optimizer with a learning rate of ``1e-3``. The
        parameters of ``transformer`` are set as optimization parameters.
    """
    return optim.Adam(transformer.parameters(), lr=1e-3)


def default_transformer_optimizer(*args: Any, **kwargs: Any) -> Any:
    msg = build_deprecation_message(
        "The function default_transformer_optimizer",
        "0.7.0",
        info="It was renamed to default_model_optimizer",
    )
    warnings.warn(msg)
    return default_model_optimizer(*args, **kwargs)


def unsupervise(image_loader: Iterable) -> Iterator[torch.Tensor]:
    for input in image_loader:
        if isinstance(input, (tuple, list)):
            input = input[0]

        yield input


def model_optimization(
    image_loader: DataLoader,
    transformer: nn.Module,
    criterion: nn.Module,
    criterion_update_fn: Optional[Callable[[torch.Tensor, nn.Module], None]] = None,
    optimizer: Optional[Optimizer] = None,
    get_optimizer: Optional[Callable[[nn.Module], Optimizer]] = None,
    quiet: bool = False,
    logger: Optional[OptimLogger] = None,
    log_fn: Optional[
        Callable[[int, Union[torch.Tensor, pystiche.LossDict], float, float], None]
    ] = None,
) -> nn.Module:
    r"""Perform a model optimization for a single epoch with integrated logging.

    Args:
        image_loader: Images used as input for the ``transformer``. Drawing from this
            should yield either an batched image or a tuple or list with a batched
            image as first item.
        transformer: Transformer to be optimized.
        criterion: Optimization criterion.
        criterion_update_fn: Is called before each optimization step with the current
            images and the optimization ``criterion``. If omitted and ``criterion`` is
            a :class:`~pystiche.loss.PerceptualLoss` or a
            :class:`~pystiche.loss.GuidedPerceptualLoss` this defaults to invoking
            :meth:`~pystiche.loss.PerceptualLoss.set_content_image`.
        optimizer: Optional optimizer. If ``None``,
            :func:`default_model_optimizer` is used.
        quiet: If ``True``, not information is logged during the optimization. Defaults
            to ``False``.
        logger: Optional custom logger. If ``None``,
            :class:`pystiche.optim.OptimLogger` is used. Defaults to ``None``.
        log_fn: Optional custom logging function. It is called in every optimization
            step with the current batch, loss as well as the image loading and
            processing velocities in img/s. If ``None``,
            :func:`pystiche.optim.default_transformer_optim_log_fn` is used. Defaults
            to ``None``.
    """
    if criterion_update_fn is None:
        if isinstance(criterion, (loss.PerceptualLoss, loss.GuidedPerceptualLoss)):

            def criterion_update_fn(  # type: ignore[misc]
                input_image: torch.Tensor,
                criterion: Union[loss.PerceptualLoss, loss.GuidedPerceptualLoss],
            ) -> None:
                criterion.set_content_image(input_image)

        else:
            raise RuntimeError(
                f"The parameter 'criterion_update_fn' can only be omitted if the "
                f"'criterion' is a loss.PerceptualLoss or a loss.GuidedPerceptualLoss. "
                f"Got {type(criterion)} instead."
            )

    if optimizer is None:
        optimizer = default_model_optimizer(transformer)

    if logger is None:
        with suppress_warnings():
            logger = OptimLogger()
    else:
        _log_parameter_deprecation("logger")

    if log_fn is None:
        with suppress_warnings():
            log_fn = default_transformer_optim_log_fn(logger, len(image_loader))
    else:
        _log_parameter_deprecation("log_fn")

    device = next(transformer.parameters()).device

    mle_handler = (
        loss.MLEHandler(criterion)
        if not isinstance(criterion, loss.MultiOperatorLoss)
        else _NullContext()
    )

    loading_time_start = time.time()
    for batch, input_image in enumerate(unsupervise(image_loader), 1):
        input_image = input_image.to(device)

        criterion_update_fn(input_image, criterion)  # type: ignore[misc]

        loading_time = time.time() - loading_time_start

        def closure() -> float:
            processing_time_start = time.time()

            # See https://github.com/pmeier/pystiche/pull/264#discussion_r430205029
            optimizer.zero_grad()  # type: ignore[union-attr]

            output_image = transformer(input_image)
            with mle_handler(output_image):  # type: ignore[operator]
                loss = criterion(output_image)
            loss.backward()

            processing_time = time.time() - processing_time_start

            if not quiet:
                batch_size = input_image.size()[0]
                image_loading_velocity = batch_size / max(loading_time, 1e-6)
                image_processing_velocity = batch_size / max(processing_time, 1e-6)
                # See https://github.com/pmeier/pystiche/pull/264#discussion_r430205029
                log_fn(batch, loss, image_loading_velocity, image_processing_velocity)  # type: ignore[misc]

            return cast(float, loss.item())

        optimizer.step(closure)
        loading_time_start = time.time()

    return transformer


def default_transformer_optim_loop(*args: Any, **kwargs: Any) -> Any:
    msg = build_deprecation_message(
        "The function default_transformer_optim_loop",
        "0.7.0",
        info="It was renamed to model_optimization",
    )
    warnings.warn(msg)
    return model_optimization(*args, **kwargs)


def multi_epoch_model_optimization(
    image_loader: DataLoader,
    transformer: nn.Module,
    criterion: nn.Module,
    criterion_update_fn: Optional[Callable[[torch.Tensor, nn.Module], None]] = None,
    epochs: int = 2,
    optimizer: Optional[Optimizer] = None,
    lr_scheduler: Optional[LRScheduler] = None,
    quiet: bool = False,
    logger: Optional[OptimLogger] = None,
    get_epoch_header: Optional[
        Callable[[int, Optimizer, Optional[LRScheduler]], str]
    ] = None,
    log_fn: Optional[
        Callable[[int, Union[torch.Tensor, pystiche.LossDict], float, float], None]
    ] = None,
) -> nn.Module:
    r"""Perform a model optimization for multiple epochs with integrated logging.

    Args:
        image_loader: Images used as input for the ``transformer``. Drawing from this
            should yield either an batched image or a tuple or list with a batched
            image as first item.
        transformer: Transformer to be optimized.
        criterion: Optimization criterion.
        criterion_update_fn: Is called before each optimization step with the current
            images and the optimization ``criterion``. If omitted and ``criterion`` is
            a :class:`~pystiche.loss.PerceptualLoss` or a
            :class:`~pystiche.loss.GuidedPerceptualLoss` this defaults to invoking
            :meth:`~pystiche.loss.PerceptualLoss.set_content_image`.
        epochs: Number of epochs. Defaults to ``2``.
        optimizer: Optional optimizer. If ``None``, it is extracted from
            ``lr_scheduler`` or func:`default_model_optimizer` is used.
        lr_scheduler: Optional learning rate scheduler. ``step()`` is invoked after
            every epoch.
        quiet: If ``True``, not information is logged during the optimization. Defaults
            to ``False``.
        logger: Optional custom logger. If ``None``,
            :class:`pystiche.optim.OptimLogger` is used. Defaults to ``None``.
        get_epoch_header: Optional custom getter for the logged epoch header. It is
        called before each epoch with the current epoch number, the ``optimizer``, and
            the ``lr_scheduler``. If ``None``
            :func:`pystiche.optim.default_epoch_header` is used. Defaults to ``None``.
        log_fn: Optional custom logging function. It is called in every optimization
            step with the current batch, loss as well as the image loading and
            processing velocities in img/s. If ``None``,
            :func:`pystiche.optim.default_transformer_optim_log_fn` is used. Defaults
            to ``None``.
        """
    if optimizer is None:
        if lr_scheduler is None:
            optimizer = default_model_optimizer(transformer)
        else:
            # Every LRScheduler has optimizer as a valid attribute
            # https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate
            # but this is not reflected in the torch type hints
            optimizer = lr_scheduler.optimizer  # type: ignore[attr-defined]

    if logger is None:
        with suppress_warnings():
            logger = OptimLogger()
    else:
        _log_parameter_deprecation("logger")

    if get_epoch_header is None:
        get_epoch_header = default_epoch_header
    else:
        _log_parameter_deprecation("get_epoch_header")

    def transformer_optim_loop(transformer: nn.Module) -> nn.Module:
        # See https://github.com/pmeier/pystiche/pull/264#discussion_r430205029
        return model_optimization(
            image_loader,
            transformer,
            criterion,
            criterion_update_fn,
            optimizer,
            quiet=quiet,
            logger=logger,
            log_fn=log_fn,
        )

    for epoch in range(epochs):
        if quiet:
            transformer = transformer_optim_loop(transformer)
        else:
            with suppress_warnings():
                header = get_epoch_header(epoch, optimizer, lr_scheduler)
            with logger.environment(header):
                transformer = transformer_optim_loop(transformer)

        if lr_scheduler is not None:
            lr_scheduler.step()

    return transformer


def default_transformer_epoch_optim_loop(*args: Any, **kwargs: Any) -> Any:
    msg = build_deprecation_message(
        "The function default_transformer_epoch_optim_loop",
        "0.7.0",
        info="It was renamed to multi_epoch_model_optimization",
    )
    warnings.warn(msg)
    return multi_epoch_model_optimization(*args, **kwargs)



from collections import OrderedDict
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union, cast

import torch
from torch import nn

import pystiche
from pystiche.enc import MultiLayerEncoder, SingleLayerEncoder
from pystiche.misc import verify_str_arg

from .op import ComparisonOperator, EncodingOperator, Operator

__all__ = [
    "OperatorContainer",
    "SameOperatorContainer",
    "MultiLayerEncodingOperator",
    "MultiRegionOperator",
]


class OperatorContainer(Operator):
    r"""Generic container for :class:`~pystiche.ops.Operator` s. If called with an image
    passes it to all immediate operators and returns a :class:`pystiche.LossDict`
    scaled with ``score_weight``.

    Args:
        named_ops: Named immediate operators that will be called if
            :class:`OperatorContainer` is called.
        score_weight: Score weight of the operator. Defaults to ``1.0``.
    """

    def __init__(
        self, named_ops: Sequence[Tuple[str, Operator]], score_weight: float = 1e0,
    ) -> None:
        super().__init__(score_weight=score_weight)
        self.add_named_modules(named_ops)

    def process_input_image(self, input_image: torch.Tensor) -> pystiche.LossDict:
        return pystiche.LossDict(
            [(name, op(input_image)) for name, op in self.named_children()]
        )

    def _get_image_or_guide(
        self, attr: str, comparison_only: bool = False
    ) -> torch.Tensor:
        images_or_guides: List[torch.Tensor] = []
        for op in self.operators():
            if comparison_only and not isinstance(op, ComparisonOperator):
                continue

            try:
                images_or_guides.append(getattr(op, attr))
            except AttributeError:
                pass

        if not images_or_guides:
            raise RuntimeError(f"No immediate children has a {attr}.")

        image_or_guide = images_or_guides[0]
        key = pystiche.TensorKey(image_or_guide)

        if not all(key == other for other in images_or_guides[1:]):
            raise RuntimeError(f"The immediate children have non-matching {attr}")

        return image_or_guide

    def get_target_guide(self) -> torch.Tensor:
        r"""Extracts the target guide from the immediate children

        Returns:
            guide: Target guide of shape :math:`1 \times 1 \times H \times W`.

        Raises:
            RuntimeError: If no immediate children has a target guide or the the target
                guides do not match each other
        """
        return self._get_image_or_guide("target_guide", comparison_only=True)

    def get_target_image(self) -> torch.Tensor:
        r"""Extracts the target image from the immediate children

        Returns:
            image: Target image of shape :math:`1 \times 1 \times H \times W`.

        Raises:
            RuntimeError: If no immediate children has a target image or the the target
                images do not match each other
        """
        return self._get_image_or_guide("target_image", comparison_only=True)

    def get_input_guide(self) -> torch.Tensor:
        r"""Extracts the input guide from the immediate children

        Returns:
            guide: Input guide of shape :math:`1 \times 1 \times H \times W`.

        Raises:
            RuntimeError: If no immediate children has a input guide or the the input
                guides do not match each other
        """
        return self._get_image_or_guide("input_guide")

    def _set_image_or_guide(
        self,
        image_or_guide: torch.Tensor,
        attr: str,
        comparison_only: bool = False,
        **kwargs: Any,
    ) -> None:
        for op in self.operators():
            if comparison_only and not isinstance(op, ComparisonOperator):
                continue

            setter = getattr(op, f"set_{attr}")
            setter(image_or_guide, **kwargs)

    def set_target_guide(self, guide: torch.Tensor, recalc_repr: bool = True) -> None:
        r"""Invoke :meth:`~pystiche.ops.ComparisonOperator.set_target_guide` on all
        immediate :class:`~pystiche.ops.ComparisonOperator` children.

        Args:
            guide: Target guide of shape :math:`1 \times 1 \times H \times W`.
            recalc_repr: If ``True``, recalculates :meth:`.target_enc_to_repr`.
                Defaults to ``True``.
        """
        self._set_image_or_guide(
            guide, "target_guide", comparison_only=True, recalc_repr=recalc_repr
        )

    def set_target_image(self, image: torch.Tensor) -> None:
        r"""Invoke :meth:`~pystiche.ops.ComparisonOperator.set_target_image` on all
        immediate :class:`~pystiche.ops.ComparisonOperator` children.

        Args:
            image: Target image of shape :math:`B \times C \times H \times W`.
        """
        self._set_image_or_guide(image, "target_image", comparison_only=True)

    def set_input_guide(self, guide: torch.Tensor) -> None:
        r"""Invoke :meth:`~pystiche.ops.Operator.set_input_guide` on all immediate
        children.

        Args:
            guide: Input guide of shape :math:`1 \times 1 \times H \times W`.
        """
        self._set_image_or_guide(guide, "input_guide")

    _modules: Dict[str, nn.Module]


class SameOperatorContainer(OperatorContainer):
    def __init__(
        self,
        names: Sequence[str],
        get_op: Callable[[str, float], Operator],
        op_weights: Union[str, Sequence[float]] = "sum",
        score_weight: float = 1e0,
    ) -> None:
        op_weights = self._parse_op_weights(op_weights, len(names))
        named_ops = [
            (name, get_op(name, weight)) for name, weight in zip(names, op_weights)
        ]

        super().__init__(named_ops, score_weight=score_weight)

    @staticmethod
    def _parse_op_weights(
        op_weights: Union[str, Sequence[float]], num_ops: int
    ) -> Sequence[float]:
        if isinstance(op_weights, str):
            verify_str_arg(op_weights, "op_weights", ("sum", "mean"))
            if op_weights == "sum":
                return [1.0] * num_ops
            else:  # op_weights == "mean":
                return [1.0 / num_ops] * num_ops
        else:
            if len(op_weights) == num_ops:
                return op_weights

            msg = (
                f"The length of the operator weights and the number of operators do "
                f"not match: {len(op_weights)} != {num_ops}"
            )
            raise ValueError(msg)


class MultiLayerEncodingOperator(SameOperatorContainer):
    r"""Convenience container for multiple :class:`~pystiche.ops.EncodingOperator` s
    operating on different ``layers`` of the same ``multi_layer_encoder``.

    Args:
        multi_layer_encoder: Multi-layer encoder.
        layers: Layers of the ``multi_layer_encoder`` that the children operators
            operate on.
        get_encoding_op: Callable that returns a children operator given a
            :class:`pystiche.enc.SingleLayerEncoder` extracted from the
            ``multi_layer_encoder`` and its corresponding layer weight.
        layer_weights: Weights of the children operators passed to ``get_encoding_op``.
            If ``"sum"``, each layer weight is set to ``1.0``. If ``"mean"``, each
            layer weight is set to ``1.0 / len(layers)``. If sequence of ``float``s its
            length has to match ``layers``. Defaults to ``"mean"``.
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Examples:

        >>> multi_layer_encoder = enc.vgg19_multi_layer_encoder()
        >>> layers = ("relu1_1", "relu2_1", "relu3_1", "relu4_1", "relu5_1")
        >>> def get_encoding_op(encoder, layer_weight):
        ...     return ops.GramOperator(encoder, score_weight=layer_weight)
        >>> op = ops.MultiLayerEncodingOperator(
        ...     multi_layer_encoder,
        ...     layers,
        ...     get_encoding_op,
        ... )
        >>> input = torch.rand(2, 3, 256, 256)
        >>> target = torch.rand(2, 3, 256, 256)
        >>> op.set_target_image(target)
        >>> score = op(input)
    """

    def __init__(
        self,
        multi_layer_encoder: MultiLayerEncoder,
        layers: Sequence[str],
        get_encoding_op: Callable[[SingleLayerEncoder, float], EncodingOperator],
        layer_weights: Union[str, Sequence[float]] = "mean",
        score_weight: float = 1e0,
    ):
        def get_op(layer: str, layer_weight: float) -> EncodingOperator:
            encoder = multi_layer_encoder.extract_encoder(layer)
            return get_encoding_op(encoder, layer_weight)

        super().__init__(
            layers, get_op, op_weights=layer_weights, score_weight=score_weight,
        )

    def __repr__(self) -> str:
        def build_encoder_repr() -> str:
            encoding_op = cast(EncodingOperator, next(self.operators()))
            multi_layer_encoder = cast(
                MultiLayerEncoder, encoding_op.encoder.multi_layer_encoder
            )
            name = multi_layer_encoder.__class__.__name__
            properties = multi_layer_encoder.properties()
            named_children = ()
            return self._build_repr(
                name=name, properties=properties, named_children=named_children
            )

        def build_op_repr(op: Operator) -> str:
            properties = op.properties()
            del properties["encoder"]
            return op._build_repr(properties=properties, named_children=())

        properties = OrderedDict()
        properties["encoder"] = build_encoder_repr()
        properties.update(self.properties())

        named_children = [
            (name, build_op_repr(op)) for name, op in self.named_operators()
        ]

        return self._build_repr(properties=properties, named_children=named_children)


class MultiRegionOperator(SameOperatorContainer):
    r"""Convenience container for multiple :class:`~pystiche.ops.Operator` s
    operating in different ``regions``.

    Args:
        regions: Regions.
        get_op: Callable that returns a children operator given a region its
            corresponding region weight.
        region_weights: Weights of the children operators passed to ``get_op``. If
            ``"sum"``, each region weight is set to ``1.0``. If ``"mean"``, each region
            weight is set to ``1.0 / len(layers)``. If sequence of ``float``s its
            length has to match ``regions``. Defaults to ``"mean"``.
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Examples:

        >>> multi_layer_encoder = enc.vgg19_multi_layer_encoder()
        >>> layers = ("relu1_1", "relu2_1", "relu3_1", "relu4_1", "relu5_1")
        >>> def get_encoding_op(encoder, layer_weight):
        ...     return ops.GramOperator(encoder, score_weight=layer_weight)
        >>> regions = ("sky", "landscape")
        >>> def get_region_op(region, region_weight):
        ...     return ops.MultiLayerEncodingOperator(
        ...         multi_layer_encoder,
        ...         layers,
        ...         get_encoding_op,
        ...         score_weight=region_weight,
        ...     )
        >>> op = ops.MultiRegionOperator(regions, get_region_op)
        >>> input = torch.rand(2, 3, 256, 256)
        >>> op.set_regional_target_image("sky", torch.rand(2, 3, 256, 256))
        >>> op.set_regional_target_image("landscape", torch.rand(2, 3, 256, 256))
        >>> score = op(input)

    """

    def __init__(
        self,
        regions: Sequence[str],
        get_op: Callable[[str, float], Operator],
        region_weights: Union[str, Sequence[float]] = "sum",
        score_weight: float = 1e0,
    ):
        super().__init__(
            regions, get_op, op_weights=region_weights, score_weight=score_weight,
        )

    def set_regional_target_guide(self, region: str, guide: torch.Tensor) -> None:
        r"""Invokes :meth:`~pystiche.ops.Comparison.set_target_guide` on the operator
        of the given ``region``.

        Args:
            region: Region.
            guide: Input guide of shape :math:`1 \times 1 \times H \times W`.
        """
        getattr(self, region).set_target_guide(guide)

    def set_regional_target_image(self, region: str, image: torch.Tensor) -> None:
        r"""Invokes :meth:`~pystiche.ops.Comparison.set_target_image` on the operator
        of the given ``region``.

        Args:
            region: Region.
            image: Input guide of shape :math:`B \times C \times H \times W`.
        """
        getattr(self, region).set_target_image(image)

    def set_regional_input_guide(self, region: str, guide: torch.Tensor) -> None:
        r"""Invokes :meth:`~pystiche.ops.Comparison.set_input_guide` on the operator
        of the given ``region``.

        Args:
            region: Region.
            guide: Input guide of shape :math:`1 \times 1 \times H \times W`.
        """
        getattr(self, region).set_input_guide(guide)



from typing import Any, Dict, Optional, Tuple, Union, cast

import torch

from pystiche.image.utils import is_edge_size, is_image_size

from . import functional as F
from .core import Transform

__all__ = ["Resize", "Rescale"]


class Resize(Transform):
    r"""Resize an image as specified

    Args:
        size: Size of the output. If ``int``, ``edge`` is used to determine the image
            size.
        edge: Corresponding edge if ``size`` is an edge size. Can be ``"short"``,
            ``"long"``, ``"vert"``, or ``"horz"``. Defaults to ``"short"``.
        aspect_ratio: Optional aspect ratio. If ``None`` the aspect ratio of ``image``
            is used. Defaults to ``None``.
        interpolation_mode: Interpolation mode used to resize ``image``. Can be
            ``"nearest"``, ``"bilinear"``, ``"bicubic"``, or ``"area"``. Defaults to
            ``"bilinear"``.
    """

    def __init__(
        self,
        size: Union[int, Tuple[int, int]],
        edge: str = "short",
        aspect_ratio: Optional[float] = None,
        interpolation_mode: str = "bilinear",
    ):
        super().__init__()
        self.size = size
        self.edge = edge
        self.aspect_ratio = aspect_ratio
        self.interpolation_mode = interpolation_mode

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return cast(
            torch.Tensor,
            F.resize(
                image,
                self.size,
                edge=self.edge,
                aspect_ratio=self.aspect_ratio,
                interpolation_mode=self.interpolation_mode,
            ),
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        if is_image_size(self.size):
            dct["image_size"] = self.size
        else:
            key = "edge_size" if is_edge_size(self.size) else "size"
            dct[key] = self.size
            dct["edge"] = self.edge
            if self.aspect_ratio is not None:
                dct["aspect_ratio"] = self.aspect_ratio
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        return dct


class Rescale(Transform):
    def __init__(
        self,
        factor: Union[float, Tuple[float, float]],
        interpolation_mode: str = "bilinear",
    ):
        super().__init__()
        self.factor = factor
        self.interpolation_mode = interpolation_mode

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.rescale(image, self.factor, self.interpolation_mode)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["factor"] = self.factor
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        return dct



import logging
import sys
import warnings
from datetime import datetime
from typing import cast

import torch
from torch import nn
from torch.nn.functional import interpolate

from pystiche.data import (
    DownloadableImage,
    DownloadableImageCollection,
    ExpiredCopyrightLicense,
    PixabayLicense,
    PublicDomainLicense,
)
from pystiche.optim import OptimLogger

from .misc import build_deprecation_message, suppress_warnings

__all__ = ["images", "logger", "transformer"]


def images() -> DownloadableImageCollection:
    """Collection of images used in the usage examples.

    .. note::

        You can use

        .. code-block:: python

            print(images())

        to get a list of all available images.
    """
    return DownloadableImageCollection(
        {
            "bird1": DownloadableImage(
                "https://download.pystiche.org/images/bird1.jpg",
                file="bird1.jpg",
                author="gholmz0",
                date="09.03.2013",
                license=PixabayLicense(),
                md5="ca869559701156c66fc1519ffc26123c",
                note="https://pixabay.com/photos/bird-wildlife-australian-bird-1139734/",
            ),
            "paint": DownloadableImage(
                "https://download.pystiche.org/images/paint.jpg",
                file="paint.jpg",
                author="garageband",
                date="03.07.2017",
                license=PixabayLicense(),
                md5="a991e222806ef49d34b172a67cf97d91",
                note="https://pixabay.com/de/photos/abstrakt-kunst-hintergrund-farbe-2468874/",
            ),
            "bird2": DownloadableImage(
                "https://download.pystiche.org/images/bird2.jpg",
                file="bird2.jpg",
                author="12019",
                date="09.04.2012",
                license=PixabayLicense(),
                md5="2398e8f6d102b9b8de7c74d4c73b71f4",
                note="https://pixabay.com/photos/bird-wildlife-sky-clouds-nature-92956/",
            ),
            "mosaic": DownloadableImage(
                "https://upload.wikimedia.org/wikipedia/commons/2/23/Mosaic_ducks_Massimo.jpg",
                file="mosaic.jpg",
                author="Marie-Lan Nguyen",
                date="2006",
                license=PublicDomainLicense(),
                md5="5b60cd1724395f7a0c21dc6dd006f8ae",
            ),
            "castle": DownloadableImage(
                "https://download.pystiche.org/images/castle.jpg",
                file="castle.jpg",
                author="Christian Hänsel",
                date="23.03.2014",
                license=PixabayLicense(),
                md5="9d085fac77065e0da66629162f759f90",
                note="https://pixabay.com/de/photos/lemgo-brake-schloss-br%C3%BCcke-310586/",
                guides=DownloadableImageCollection(
                    {
                        "building": DownloadableImage(
                            "https://download.pystiche.org/images/castle/building.png",
                            file="building.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="680bd875f4fdb919d446d6210cbe1035",
                        ),
                        "sky": DownloadableImage(
                            "https://download.pystiche.org/images/castle/sky.png",
                            file="sky.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="c4d5f18b3ca22a1ff879f40ca797ed26",
                        ),
                        "water": DownloadableImage(
                            "https://download.pystiche.org/images/castle/water.png",
                            file="water.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="2642552e8c98b6a63f24dd67bb54e2a3",
                        ),
                    },
                ),
            ),
            "church": DownloadableImage(
                "https://upload.wikimedia.org/wikipedia/commons/a/ad/Vincent_van_Gogh_-_The_Church_in_Auvers-sur-Oise%2C_View_from_the_Chevet_-_Google_Art_Project.jpg",
                file="church.jpg",
                author="Vincent van Gogh",
                title="The Church at Auvers",
                date="1890",
                license=ExpiredCopyrightLicense(1890),
                md5="fd866289498367afd72a5c9cf626073a",
                guides=DownloadableImageCollection(
                    {
                        "building": DownloadableImage(
                            "https://download.pystiche.org/images/church/building.png",
                            file="building.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="41256f62d2bd1560ed1e66623c8d9c9f",
                        ),
                        "sky": DownloadableImage(
                            "https://download.pystiche.org/images/church/sky.png",
                            file="sky.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="c87c7234d2788a1d2a4e1633723c794b",
                        ),
                        "surroundings": DownloadableImage(
                            "https://download.pystiche.org/images/church/surroundings.png",
                            file="surroundings.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="807a40beed727af81333edbd8eb89aff",
                        ),
                    }
                ),
            ),
            "cliff": DownloadableImage(
                "https://upload.wikimedia.org/wikipedia/commons/a/a4/Claude_Monet_-_Cliff_Walk_at_Pourville_-_Google_Art_Project.jpg",
                file="cliff.jpg",
                author="Claude Monet",
                title="The Cliff Walk at Pourville",
                date="1882",
                license=ExpiredCopyrightLicense(1926),
                md5="8f6b8101b484f17cea92a12ad27be31d",
                guides=DownloadableImageCollection(
                    {
                        "landscape": DownloadableImage(
                            "https://download.pystiche.org/images/cliff/landscape.png",
                            file="landscape.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="4d02ee8aaa610ced3ef12a5c82f29b81",
                        ),
                        "sky": DownloadableImage(
                            "https://download.pystiche.org/images/cliff/sky.png",
                            file="sky.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="8f76b7ed88ff1b23feaab1de51501ae1",
                        ),
                        "water": DownloadableImage(
                            "https://download.pystiche.org/images/cliff/water.png",
                            file="water.png",
                            author="Julian Bültemeier",
                            license=PublicDomainLicense(),
                            md5="f0cab02afadf38d3262ce1c61a36f6da",
                        ),
                    }
                ),
            ),
        }
    )


def demo_images() -> DownloadableImageCollection:
    warnings.warn(
        build_deprecation_message(
            "The function demo_images", "0.6.0", info="It was renamed to images."
        )
    )
    return images()


def logger() -> OptimLogger:
    """Simple logger used in the usage examples.
    """
    msg = build_deprecation_message(
        "The function logger",
        "0.7.0",
        "See https://github.com/pmeier/pystiche/issues/434 for details.",
    )
    warnings.warn(msg, UserWarning)
    logger = logging.getLogger(
        name=f"pystiche_{datetime.now().strftime('%Y%m%d%H%M%S%f')}"
    )
    logger.setLevel(logging.INFO)

    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(logging.INFO)
    logger.addHandler(sh)

    with suppress_warnings():
        return OptimLogger(logger)


def demo_logger() -> OptimLogger:
    warnings.warn(
        build_deprecation_message(
            "The function demo_logger", "0.6.0", info="It was renamed to logger."
        )
    )
    return logger()


class Interpolate(nn.Module):
    def __init__(self, scale_factor: float = 1.0, mode: str = "nearest") -> None:
        super().__init__()
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return cast(
            torch.Tensor,
            interpolate(input, scale_factor=self.scale_factor, mode=self.mode),
        )

    def extra_repr(self) -> str:
        extras = [f"scale_factor={self.scale_factor}"]
        if self.mode != "nearest":
            extras.append(f"mode={self.mode}")
        return ", ".join(extras)


class Conv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        *,
        stride: int = 1,
        upsample: bool = False,
        norm: bool = True,
        activation: bool = True,
    ):
        super().__init__()
        self.upsample = Interpolate(scale_factor=stride) if upsample else None
        self.pad = nn.ReflectionPad2d(kernel_size // 2)
        self.conv = nn.Conv2d(
            in_channels, out_channels, kernel_size, stride=1 if upsample else stride
        )
        self.norm = nn.InstanceNorm2d(out_channels, affine=True) if norm else None
        self.activation = nn.ReLU() if activation else None

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.upsample:
            input = self.upsample(input)

        output = self.conv(self.pad(input))

        if self.norm:
            output = self.norm(output)
        if self.activation:
            output = self.activation(output)

        return cast(torch.Tensor, output)


class Residual(nn.Module):
    def __init__(self, channels: int) -> None:
        super().__init__()
        self.conv1 = Conv(channels, channels, kernel_size=3)
        self.conv2 = Conv(channels, channels, kernel_size=3, activation=False)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        output = cast(torch.Tensor, self.conv2(self.conv1(input)))
        return output + input


class FloatToUint8Range(nn.Module):
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return input * 255.0


class Uint8ToFloatRange(nn.Module):
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return input / 255.0


class Transformer(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.encoder = nn.Sequential(
            Conv(3, 32, kernel_size=9),
            Conv(32, 64, kernel_size=3, stride=2),
            Conv(64, 128, kernel_size=3, stride=2),
            Residual(128),
            Residual(128),
            Residual(128),
            Residual(128),
            Residual(128),
        )
        self.decoder = nn.Sequential(
            Conv(128, 64, kernel_size=3, stride=2, upsample=True),
            Conv(64, 32, kernel_size=3, stride=2, upsample=True),
            Conv(32, 3, kernel_size=9, norm=False, activation=False),
        )

        self.preprocessor = FloatToUint8Range()
        self.postprocessor = Uint8ToFloatRange()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        input = self.preprocessor(input)
        output = self.decoder(self.encoder(input))
        return cast(torch.Tensor, self.postprocessor(output))


def transformer() -> nn.Module:
    """Basic transformer for model-based optimization.

    The transformer is compatible with the
    `official PyTorch example <https://github.com/pytorch/examples/tree/master/fast_neural_style>`_
    which in turn is based on :cite:`JAL2016`
    """
    return Transformer()



from typing import cast

import torch
import torch.nn.functional as F
from torch import nn

from pystiche.meta import (
    ConvModule,
    PoolModule,
    conv_module_meta,
    is_conv_module,
    is_pool_module,
    pool_module_meta,
)
from pystiche.misc import verify_str_arg

__all__ = ["propagate_guide"]


def propagate_guide(
    module: nn.Module,
    guide: torch.Tensor,
    method: str = "simple",
    allow_empty: bool = False,
) -> torch.Tensor:
    verify_str_arg(method, "method", ("simple", "inside", "all"))
    if is_conv_module(module):
        guide = _conv_guide(cast(ConvModule, module), guide, method)
    elif is_pool_module(module):
        guide = _pool_guide(cast(PoolModule, module), guide)

    if allow_empty or torch.any(guide.bool()):
        return guide

    msg = (
        f"Guide has no longer any entries after propagation through "
        f"{module.__class__.__name__}({module.extra_repr()}). If this is valid, "
        f"set allow_empty=True."
    )
    raise RuntimeError(msg)


def _conv_guide(module: ConvModule, guide: torch.Tensor, method: str) -> torch.Tensor:
    # TODO: deal with convolution that doesn't preserve the output shape
    if method == "simple":
        return guide

    meta = conv_module_meta(module)
    guide_unfolded = F.unfold(guide, **meta).byte()

    if method == "inside":
        mask = ~torch.all(guide_unfolded, 1, keepdim=True)
        val = False
    else:
        mask = torch.any(guide_unfolded, 1, keepdim=True)
        val = True

    mask, _ = torch.broadcast_tensors(mask, guide_unfolded)
    guide_unfolded[mask] = val

    guide_folded = F.fold(guide_unfolded.float(), guide.size()[2:], **meta)
    return torch.clamp(guide_folded, 0.0, 1.0)


def _pool_guide(module: PoolModule, guide: torch.Tensor) -> torch.Tensor:
    return cast(torch.Tensor, F.max_pool2d(guide, **pool_module_meta(module)))



import warnings
from collections import OrderedDict
from copy import copy
from typing import Collection, Dict, Iterator, Optional, Sequence, Set, Tuple, cast

import torch
from torch import nn

import pystiche
from pystiche.misc import suppress_warnings

from .encoder import Encoder
from .guides import propagate_guide

__all__ = ["MultiLayerEncoder", "SingleLayerEncoder"]


def _future_warning(name: str) -> None:
    msg = (
        f"The functionality of MultiLayerEncoder.{name} will change in the future. "
        f"If you depend on this functionality, "
        f"see https://github.com/pmeier/pystiche/issues/435 for details "
    )
    warnings.warn(msg, FutureWarning)


class MultiLayerEncoder(pystiche.Module):
    r"""Sequential architecture that supports extracting encodings from intermediate
    layers in a single forward pass. Encodings can be stored to avoid recalculating
    them if they are required multiple times. Invokes :meth:`MultiLayerEncoder.forward`
    if called.

    Args:
        modules: Named sequential modules.

    Attributes:
        registered_layers: Set of registered layers for
            :meth:`MultiLayerEncoder.encode` and :meth:`MultiLayerEncoder.trim`.
    """

    def __init__(self, modules: Sequence[Tuple[str, nn.Module]]) -> None:
        super().__init__(named_children=modules)
        self.registered_layers: Set[str] = set()
        self._storage: Dict[Tuple[str, pystiche.TensorKey], torch.Tensor] = dict()

        # TODO: remove this?
        self.requires_grad_(False)
        self.eval()

    def children_names(self) -> Iterator[str]:
        for name, child in self.named_children():
            yield name

    def __contains__(self, layer: str) -> bool:
        r"""Checks if the given layer is part of the :class:`MultiLayerEncoder`

        Args:
            layer: Layer.
        """
        return layer in self.children_names()

    def _verify_layer(self, layer: str) -> None:
        if layer not in self:
            raise ValueError(f"Layer {layer} is not part of the encoder.")

    def extract_deepest_layer(self, layers: Collection[str]) -> str:
        for layer in layers:
            self._verify_layer(layer)
        return sorted(set(layers), key=list(self.children_names()).index)[-1]

    def named_children_to(
        self, layer: str, include_last: bool = False
    ) -> Iterator[Tuple[str, nn.Module]]:
        self._verify_layer(layer)
        idx = list(self.children_names()).index(layer)
        if include_last:
            idx += 1
        for name, child in tuple(self.named_children())[:idx]:
            yield name, child

    def named_children_from(
        self, layer: str, include_first: bool = True
    ) -> Iterator[Tuple[str, nn.Module]]:
        self._verify_layer(layer)
        idx = list(self.children_names()).index(layer)
        if not include_first:
            idx += 1
        for name, child in tuple(self.named_children())[idx:]:
            yield name, child

    def forward(
        self, input: torch.Tensor, layers: Sequence[str], store: bool = False
    ) -> Tuple[torch.Tensor, ...]:
        r"""Encode the input on the given layers in a single forward pass. If the input
        was encoded before the encodings are extracted from the storage rather than
        executing the forward pass again.

        Args:
            input: Input.
            layers: Layers.
            store: If ``True``, store the encodings.

        Returns:
            Tuple of encodings which order corresponds to ``layers``.
        """
        _future_warning("__call__")
        storage = copy(self._storage)
        input_key = pystiche.TensorKey(input)
        stored_layers = [name for name, key in storage.keys() if key == input_key]
        diff_layers = set(layers) - set(stored_layers)

        if diff_layers:
            deepest_layer = self.extract_deepest_layer(diff_layers)
            for name, module in self.named_children_to(
                deepest_layer, include_last=True
            ):
                input = storage[(name, input_key)] = module(input)

            if store:
                self._storage = storage

        return tuple(storage[(name, input_key)] for name in layers)

    def extract_encoder(self, layer: str) -> "SingleLayerEncoder":
        r"""Extract a :class:`SingleLayerEncoder` for the given layer and register
        the layer in :attr:`MultiLayerEncoder.registered_layers`.

        Args:
            layer: Layer.
        """
        self._verify_layer(layer)
        self.registered_layers.add(layer)
        return SingleLayerEncoder(self, layer)

    def encode(self, input: torch.Tensor) -> None:
        r"""Encode the given input and store the encodings of all
        :attr:`MultiLayerEncoder.registered_layers`.

        Args:
            input: Input.
        """
        _future_warning("encode")

        if not self.registered_layers:
            return

        key = pystiche.TensorKey(input)
        keys = [(layer, key) for layer in self.registered_layers]
        encs = self(input, layers=self.registered_layers, store=True)
        self._storage = dict(zip(keys, encs))

    def empty_storage(self) -> None:
        r"""Empty the encodings storage."""
        self._storage = {}

    def trim(self, layers: Optional[Collection[str]] = None) -> None:
        r"""Remove excess layers that are not necessary to generate the encodings of
        ``layers``.

        Args:
            layers: Layers that the :class:`MultiLayerEncoder` need to be able to
                generate encodings for. If ``None``,
                :attr:`MultiLayerEncoder.registered_layers` is used. Defaults to
                ``None``.
        """
        if layers is None:
            layers = self.registered_layers
        deepest_layer = self.extract_deepest_layer(layers)
        for name, _ in self.named_children_from(deepest_layer, include_first=False):
            del self._modules[name]

    def propagate_guide(
        self,
        guide: torch.Tensor,
        layers: Sequence[str],
        method: str = "simple",
        allow_empty: bool = False,
    ) -> Tuple[torch.Tensor, ...]:
        r"""Propagate the guide on the given layers.

        Args:
            guide: Guide.
            layers: Layers.
            allow_empty: If ``True``, allow the propagated guides to become empty.
                Defaults to ``False``.

        Returns:
            Tuple of guides which order corresponds to ``layers``.
        """
        guides = {}
        deepest_layer = self.extract_deepest_layer(layers)
        for name, module in self.named_children_to(deepest_layer, include_last=True):
            try:
                guide = guides[name] = propagate_guide(
                    module, guide, method=method, allow_empty=allow_empty
                )
            except RuntimeError as error:
                # TODO: customize error message to better reflect which layer causes
                #       the problem
                raise error

        return tuple(guides[name] for name in layers)


class SingleLayerEncoder(Encoder):
    r"""Encoder extracted from a :class:`MultiLayerEncoder` that operates on a single
    layer. Invokes :meth:`SingleLayerEncoder.forward` if called.

    Attributes:
        multi_layer_encoder: Corresponding multi-layer encoder.
        layer: Encoding layer.
    """

    def __init__(self, multi_layer_encoder: MultiLayerEncoder, layer: str):
        super().__init__()
        self.multi_layer_encoder = multi_layer_encoder
        self.layer = layer

    def forward(self, input_image: torch.Tensor) -> torch.Tensor:
        r"""Encode the given input image on :attr:`SingleLayerEncoder.layer` of
        :attr:`SingleLayerEncoder.multi_layer_encoder`.

        Args:
            input_image: Input image.
        """
        with suppress_warnings(FutureWarning):
            return cast(
                Tuple[torch.Tensor],
                self.multi_layer_encoder(input_image, layers=(self.layer,)),
            )[0]

    def propagate_guide(self, guide: torch.Tensor) -> torch.Tensor:
        r"""Propagate the given guide on :attr:`SingleLayerEncoder.layer` of
        :attr:`SingleLayerEncoder.multi_layer_encoder`.

        Args:
            guide: Guide.
        """
        return self.multi_layer_encoder.propagate_guide(guide, layers=(self.layer,))[0]

    def __repr__(self) -> str:
        name = self.multi_layer_encoder.__class__.__name__
        properties = OrderedDict()
        properties["layer"] = self.layer
        properties.update(self.multi_layer_encoder.properties())
        named_children = ()
        return self._build_repr(
            name=name, properties=properties, named_children=named_children
        )



from os import path
from typing import Any, Callable, Dict, Iterator, Mapping, Optional, Tuple, TypeVar

import torch

import pystiche
from pystiche.image import read_image

_ImageCollection_co = TypeVar("_ImageCollection_co", covariant=True)


class _Image(pystiche.ComplexObject):
    def __init__(
        self,
        file: str,
        guides: Optional["_ImageCollection"] = None,
        transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        note: Optional[str] = None,
    ) -> None:
        self.file = file
        self.guides = guides
        self.transform = transform
        self.note = note

    def read(
        self, root: Optional[str] = None, **read_image_kwargs: Any,
    ) -> torch.Tensor:
        r"""Read the image from file with :func:`pystiche.image.read_image` and
        optionally apply :attr:`.transform`.

        Args:
            root: Optional root directory if the file is a relative path.
            **read_image_kwargs: Optional parameters passed to
                :func:`pystiche.image.read_image`.
        """
        file = self.file
        if not path.isabs(file) and root is not None:
            file = path.join(root, file)

        image = read_image(file, **read_image_kwargs)
        if self.transform is None:
            return image

        return self.transform(image)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["file"] = self.file
        if self.transform is not None:
            dct["transform"] = self.transform
        if self.note is not None:
            dct["note"] = self.note
        return dct

    def _named_children(self) -> Iterator[Tuple[str, Any]]:
        yield from super()._named_children()
        if self.guides is not None:
            yield from self.guides


class _ImageCollection(pystiche.ComplexObject):
    def __init__(self, images: Mapping[str, _Image]) -> None:
        self._images = images

    def __len__(self) -> int:
        return len(self._images)

    def __getitem__(self, name: str) -> _Image:
        return self._images[name]

    def __iter__(self) -> Iterator[Tuple[str, _Image]]:
        for name, image in self._images.items():
            yield name, image

    def read(
        self, root: Optional[str] = None, **read_image_kwargs: Any
    ) -> Dict[str, torch.Tensor]:
        r"""Read the images from file.

        Args:
            root: Optional root directory if the file is a relative path.
            **read_image_kwargs: Optional parameters passed to
                :func:`pystiche.image.read_image`.

        Returns:
            Dictionary with the name image pairs.
        """
        return {
            name: image.read(root=root, **read_image_kwargs) for name, image in self
        }

    def _named_children(self) -> Iterator[Tuple[str, Any]]:
        yield from iter(self._images.items())



import os
from os import path
from typing import Dict, Optional

from torch import nn
from torchvision.datasets.folder import is_image_file

from ._core import _Image, _ImageCollection


class LocalImage(_Image):
    def __init__(
        self,
        file: str,
        collect_local_guides: bool = True,
        guides: Optional[_ImageCollection] = None,
        transform: Optional[nn.Module] = None,
        note: Optional[str] = None,
    ):
        if collect_local_guides and guides is None:
            guides = self._collect_guides(file)
        super().__init__(file, guides=guides, transform=transform, note=note)

    @staticmethod
    def _collect_guides(file: str) -> Optional["LocalImageCollection"]:
        dir = path.splitext(file)[0]
        if not path.isdir(dir):
            return None

        image_files = [file for file in os.listdir(dir) if is_image_file(file)]
        if not image_files:
            return None

        guides: Dict[str, "LocalImage"] = {}
        for file in image_files:
            region = path.splitext(path.basename(file))[0]
            guides[region] = LocalImage(
                path.join(dir, file), collect_local_guides=False
            )
        return LocalImageCollection(guides)


class LocalImageCollection(_ImageCollection):
    pass



from typing import List, Optional, Tuple, Union, cast

import torch
from torch.nn.functional import affine_grid as _affine_grid
from torch.nn.functional import grid_sample as _grid_sample
from torch.nn.functional import interpolate as _interpolate

__all__ = ["interpolate", "affine_grid", "grid_sample"]

# The purpose of this module is to suppress UserWarnings from interpolate(),
# grid_sample(), and affine_grid() due to the changed default value of grid_sample


def _get_align_corners(interpolation_mode: str) -> Optional[bool]:
    if interpolation_mode in ("nearest", "area"):
        return None
    else:
        return False


def interpolate(
    image: torch.Tensor,
    size: Optional[Union[int, Tuple[int, int]]] = None,
    scale_factor: Optional[float] = None,
    mode: str = "nearest",
) -> torch.Tensor:
    return cast(
        torch.Tensor,
        _interpolate(
            image,
            size=size,
            scale_factor=scale_factor,
            mode=mode,
            align_corners=_get_align_corners(mode),
        ),
    )


def affine_grid(theta: torch.Tensor, size: Tuple[int, int, int, int],) -> torch.Tensor:
    return _affine_grid(theta, cast(List[int], size), align_corners=False)


def grid_sample(
    image: torch.Tensor,
    grid: torch.Tensor,
    mode: str = "bilinear",
    padding_mode: str = "zeros",
) -> torch.Tensor:
    return _grid_sample(
        image,
        grid,
        mode=mode,
        padding_mode=padding_mode,
        align_corners=_get_align_corners(mode),
    )



from torch import nn

from pystiche.image.transforms import CaffePreprocessing, TorchPreprocessing
from pystiche.misc import suppress_warnings

PREPROCESSORS = {"torch": TorchPreprocessing, "caffe": CaffePreprocessing}

__all__ = ["get_preprocessor"]


def get_preprocessor(framework: str) -> nn.Module:
    with suppress_warnings():
        return PREPROCESSORS[framework]()



import itertools
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np

import torch

import pystiche
from pystiche.enc import Encoder
from pystiche.image.transforms import Transform, TransformMotifAffinely
from pystiche.misc import suppress_warnings, to_2d_arg

from . import functional as F
from .op import EncodingComparisonOperator

__all__ = [
    "FeatureReconstructionOperator",
    "GramOperator",
    "MRFOperator",
]


class FeatureReconstructionOperator(EncodingComparisonOperator):
    r"""The feature reconstruction loss is the de facto standard content loss. It
    measures the mean squared error (MSE) between the encodings of an ``input_image``
    :math:`\hat{I}` and a ``target_image`` :math:`I` :

    .. math::

        \mean \parentheses{\Phi\of{\hat{I}} - \Phi\of{I}}^2

    Here :math:`\Phi\of{\cdot}` denotes the ``encoder``.

    .. note::

        Opposed to the paper, the implementation calculates the grand average
        :math:`\mean` opposed to the grand sum :math:`\sum` to account for different
        sized images.

    Args:
        encoder: Encoder :math:`\Phi`.
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Examples:

        >>> multi_layer_encoder = enc.vgg19_multi_layer_encoder()
        >>> encoder = multi_layer_encoder.extract_encoder("relu4_2")
        >>> op = ops.FeatureReconstructionOperator(encoder)
        >>> input = torch.rand(2, 3, 256, 256)
        >>> target = torch.rand(2, 3, 256, 256)
        >>> op.set_target_image(target)
        >>> score = op(input)

    .. seealso::

        The feature reconstruction loss was introduced by Mahendran and Vedaldi in
        :cite:`MV2015` , but its name was coined by Johnson, Alahi, and Fei-Fei in
        :cite:`JAFF2016` .
    """

    def enc_to_repr(self, enc: torch.Tensor) -> torch.Tensor:
        return enc

    def input_enc_to_repr(
        self, enc: torch.Tensor, ctx: Optional[torch.Tensor]
    ) -> torch.Tensor:
        return self.enc_to_repr(enc)

    def target_enc_to_repr(self, enc: torch.Tensor) -> Tuple[torch.Tensor, None]:
        return self.enc_to_repr(enc), None

    def calculate_score(
        self,
        input_repr: torch.Tensor,
        target_repr: torch.Tensor,
        ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        return F.mse_loss(input_repr, target_repr)


class GramOperator(EncodingComparisonOperator):
    r"""The gram loss is a style loss based on the correlation of feature map channels.
    It measures the mean squared error (MSE) between the channel-wise Gram matrices of
    the encodings of an ``input_image`` :math:`\hat{I}` and a ``target_image``
    :math:`I`:

    .. math::

        \mean \parentheses{\fun{gram}{\Phi\of{\hat{I}}} - \fun{gram}{\Phi\of{I}}}^2

    Here :math:`\Phi\of{\cdot}` denotes the ``encoder`` and :math:`\fun{gram}{\cdot}`
    denotes :func:`pystiche.gram_matrix`.

    .. note::

        Opposed to the paper, the implementation calculates the grand average
        :math:`\mean` opposed to the grand sum :math:`\sum` to account for different
        sized images.

    Args:
        encoder: Encoder :math:`\Phi\left( \cdot \right)`.
        normalize: If ``True``, normalizes the Gram matrices to account for different
            sized images. See :func:`pystiche.gram_matrix` for details. Defaults to
            ``True``.
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Examples:

        >>> multi_layer_encoder = enc.vgg19_multi_layer_encoder()
        >>> encoder = multi_layer_encoder.extract_encoder("relu4_2")
        >>> op = ops.GramOperator(encoder)
        >>> input = torch.rand(2, 3, 256, 256)
        >>> target = torch.rand(2, 3, 256, 256)
        >>> op.set_target_image(target)
        >>> score = op(input)

    .. seealso::

        The feature reconstruction loss was introduced by Gatys, Ecker, and Bethge
        in :cite:`GEB2016` .
    """

    def __init__(
        self, encoder: Encoder, normalize: bool = True, score_weight: float = 1.0
    ) -> None:
        super().__init__(encoder, score_weight=score_weight)
        self.normalize = normalize

    def enc_to_repr(self, enc: torch.Tensor) -> torch.Tensor:
        return pystiche.gram_matrix(enc, normalize=self.normalize)

    def input_enc_to_repr(
        self, enc: torch.Tensor, ctx: Optional[torch.Tensor]
    ) -> torch.Tensor:
        return self.enc_to_repr(enc)

    def target_enc_to_repr(self, enc: torch.Tensor) -> Tuple[torch.Tensor, None]:
        return self.enc_to_repr(enc), None

    def calculate_score(
        self,
        input_repr: torch.Tensor,
        target_repr: torch.Tensor,
        ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        return F.mse_loss(input_repr, target_repr)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        if not self.normalize:
            dct["normalize"] = self.normalize
        return dct


class MRFOperator(EncodingComparisonOperator):
    r"""The MRF loss is a style loss based on
    `Markov Random Fields (MRFs) <https://en.wikipedia.org/wiki/Markov_random_field>`_.
    It measures the mean squared error (MSE) between *neural patches* extracted from
    the encodings of an ``input_image`` :math:`\hat{I}` and a ``target_image``
    :math:`I`:

    .. math::

        \mean \parentheses{p_n\of{\Phi\of{\hat{I}}} - p_{MCS\of{n}}\of{\Phi\of{\hat{I}}}}^2

    Since the number of patches might be different for both images and the order of the
    patches does not correlate with the order of the enclosed style element, for each
    input neural patch :math:`n` a fitting target neural patch is to selected based on
    the maximum cosine similarity :math:`MCS\of{n}` with
    :func:`pystiche.cosine_similarity`.

    .. note::

        Opposed to the paper, the implementation calculates the grand average
        :math:`\mean` opposed to the grand sum :math:`\sum` to account for different
        sized images.

    Args:
        encoder: Encoder :math:`\Phi` .
        patch_size: Spatial size of the neural patches.
        stride: Distance between two neural patches.
        target_transforms: Optional transformations to apply to the target image before
            the neural patches are extracted. Defaults to ``None``.
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Examples:

        >>> multi_layer_encoder = enc.vgg19_multi_layer_encoder()
        >>> encoder = multi_layer_encoder.extract_encoder("relu4_2")
        >>> patch_size = 3
        >>> op = ops.MRFOperator(encoder, patch_size)
        >>> input = torch.rand(2, 3, 256, 256)
        >>> target = torch.rand(2, 3, 256, 256)
        >>> op.set_target_image(target)
        >>> score = op(input)

    .. seealso::

        The MRF loss was introduced by Li and Wand in :cite:`LW2016`.
    """

    def __init__(
        self,
        encoder: Encoder,
        patch_size: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        target_transforms: Optional[Iterable[Transform]] = None,
        score_weight: float = 1.0,
    ):
        super().__init__(encoder, score_weight=score_weight)
        self.patch_size = to_2d_arg(patch_size)
        self.stride = to_2d_arg(stride)
        self.target_transforms = target_transforms

    @staticmethod
    def scale_and_rotate_transforms(
        num_scale_steps: int = 1,
        scale_step_width: float = 5e-2,
        num_rotate_steps: int = 1,
        rotate_step_width: float = 10.0,
    ) -> List[TransformMotifAffinely]:
        """Generate a list of scaling and rotations transformations.

        .. seealso::

            The output of this method can be used as parameter ``target_transforms`` of
            :class:`~pystiche.ops.MRFOperator` to enrich the space of target neural
            patches:

            .. code-block:: python

                from pystiche.ops import MRFOperator


                target_transforms = MRFOperator.scale_and_rotate_transforms()
                op = MRFOperator(..., target_transforms=target_transforms)

        Args:
            num_scale_steps: Number of scale steps. Each scale is performed in both
                directions, i.e. enlarging and shrinking the motif. Defaults to ``1``.
            scale_step_width: Width of each scale step. Defaults to ``5e-2``.
            num_rotate_steps: Number of rotate steps. Each rotate is performed in both
                directions, i.e. clockwise and counterclockwise. Defaults to ``1``.
            rotate_step_width: Width of each rotation step in degrees.
                Defaults to ``10.0``.

        Returns:
           ``(num_scale_steps * 2 + 1) * (num_rotate_steps * 2 + 1)`` transformations
           in total comprising every combination given by the input parameters.
        """
        scaling_factors = np.arange(
            -num_scale_steps, num_scale_steps + 1, dtype=np.float
        )
        scaling_factors = 1.0 + (scaling_factors * scale_step_width)

        rotation_angles = np.arange(
            -num_rotate_steps, num_rotate_steps + 1, dtype=np.float
        )
        rotation_angles *= rotate_step_width

        transforms = []
        for scaling_factor, rotation_angle in itertools.product(
            scaling_factors, rotation_angles
        ):
            with suppress_warnings():
                transform = TransformMotifAffinely(
                    scaling_factor=scaling_factor,
                    rotation_angle=rotation_angle,
                    canvas="same",  # FIXME: this should be valid after it is implemented
                )
            transforms.append(transform)

        return transforms

    @staticmethod
    def _match_batch_sizes(target: torch.Tensor, input: torch.Tensor) -> torch.Tensor:
        # FIXME
        return target

    def set_target_guide(self, guide: torch.Tensor, recalc_repr: bool = True) -> None:
        # Since the target representation of the MRFOperator possibly comprises
        # scaled or rotated patches, it is not useful to store the target encoding
        # guides
        self.register_buffer("target_guide", guide)
        if recalc_repr and self.has_target_image:
            self.set_target_image(self.target_image)

    def _guide_repr(self, repr: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
        # Due to the guiding large areas of the images might be zero and thus many
        # patches might carry no information. These patches can be removed from the
        # target and input representation reducing the computing cost and memory during
        # the loss calculation.

        # Patches without information have constant values in the spatial dimensions.
        repr_flat = torch.flatten(repr, 2)
        constant = repr_flat[:, :, 0].unsqueeze(2)

        # By checking where the spatial values do not differ from this constant in any
        # channel, the patches with no information can be filtered out.
        abs_diff = torch.abs(repr_flat - constant)
        mask = torch.any(torch.flatten(abs_diff > eps, 1), dim=1)

        return repr[mask]

    def enc_to_repr(self, enc: torch.Tensor, is_guided: bool) -> torch.Tensor:
        with suppress_warnings(FutureWarning):
            repr = pystiche.extract_patches2d(enc, self.patch_size, self.stride)
        if not is_guided:
            return repr

        return self._guide_repr(repr)

    def input_enc_to_repr(
        self, enc: torch.Tensor, ctx: Optional[torch.Tensor]
    ) -> torch.Tensor:
        return self.enc_to_repr(enc, self.has_input_guide)

    def target_enc_to_repr(self, enc: torch.Tensor) -> Tuple[torch.Tensor, None]:
        return self.enc_to_repr(enc, self.has_target_guide), None

    def target_image_to_repr(self, image: torch.Tensor) -> Tuple[torch.Tensor, None]:
        # Due to the possible scaling and rotation, we only apply the guide to the
        # target image and not the encodings
        if self.has_target_guide:
            image = self.apply_guide(image, self.target_guide)

        if self.target_transforms is None:
            return self.target_enc_to_repr(self.encoder(image))

        device = image.device
        reprs = []
        for transform in self.target_transforms:
            transform = transform.to(device)
            enc = self.encoder(transform(image))
            repr, _ = self.target_enc_to_repr(enc)
            reprs.append(repr)

        repr = torch.cat(reprs)
        ctx = None
        return repr, ctx

    def calculate_score(
        self,
        input_repr: torch.Tensor,
        target_repr: torch.Tensor,
        ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        return F.mrf_loss(input_repr, target_repr, batched_input=False)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["patch_size"] = self.patch_size
        dct["stride"] = self.stride
        return dct



import warnings
from typing import Sequence, Union

import torch

from pystiche.misc import prod, to_1d_arg, to_2d_arg, to_3d_arg, zip_equal

__all__ = [
    "extract_patches1d",
    "extract_patches2d",
    "extract_patches3d",
]


def _warn_output_shape(*dim_names: str) -> None:
    name = f"extract_patches{len(dim_names)}d"
    partial_shape = "x".join(dim_names)
    warnings.warn(
        f"The output shape of {name} will change in the future. "
        f"The current shape B*PxCx{partial_shape} will be replaced by "
        f"BxPxCx{partial_shape} thus adding a dimension. "
        f"Here, 'P' denotes the number of extracted patches.",
        FutureWarning,
    )


def _extract_patchesnd(
    x: torch.Tensor, patch_sizes: Sequence[int], strides: Sequence[int]
) -> torch.Tensor:
    num_channels = x.size()[1]
    dims = range(2, x.dim())
    for dim, patch_size, stride in zip_equal(dims, patch_sizes, strides):
        x = x.unfold(dim, patch_size, stride)
    x = x.permute(0, *dims, 1, *[dim + len(dims) for dim in dims]).contiguous()
    num_patches = prod(x.size()[: len(dims) + 1])
    return x.view(num_patches, num_channels, *patch_sizes)


def extract_patches1d(
    x: torch.Tensor,
    patch_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]] = 1,
) -> torch.Tensor:
    assert x.dim() == 3
    _warn_output_shape("L")
    return _extract_patchesnd(x, to_1d_arg(patch_size), to_1d_arg(stride))


def extract_patches2d(
    x: torch.Tensor,
    patch_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]] = 1,
) -> torch.Tensor:
    assert x.dim() == 4
    _warn_output_shape("H", "W")
    return _extract_patchesnd(x, to_2d_arg(patch_size), to_2d_arg(stride))


def extract_patches3d(
    x: torch.Tensor,
    patch_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]] = 1,
) -> torch.Tensor:
    assert x.dim() == 5
    _warn_output_shape("D", "H", "W")
    return _extract_patchesnd(x, to_3d_arg(patch_size), to_3d_arg(stride))



from typing import cast

import torch

from . import functional as F
from .core import Transform

__all__ = [
    "RGBToGrayscale",
    "GrayscaleToFakegrayscale",
    "RGBToFakegrayscale",
    "GrayscaleToBinary",
    "RGBToBinary",
    "RGBToYUV",
    "YUVToRGB",
]


class RGBToGrayscale(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.rgb_to_grayscale(x)


class GrayscaleToFakegrayscale(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, F.grayscale_to_fakegrayscale(x))


class RGBToFakegrayscale(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.rgb_to_fakegrayscale(x)


class GrayscaleToBinary(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.grayscale_to_binary(x)


class RGBToBinary(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.rgb_to_binary(x)


class RGBToYUV(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.rgb_to_yuv(x)


class YUVToRGB(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.yuv_to_rgb(x)



import contextlib
import itertools
import warnings
from functools import reduce as _reduce
from operator import mul
from os import path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.error import HTTPError
from urllib.request import Request, urlopen

import torch
from torchvision.datasets.utils import check_md5

__all__ = [
    "prod",
    "to_1d_arg",
    "to_2d_arg",
    "to_3d_arg",
    "zip_equal",
    "build_fmtstr",
    "verify_str_arg",
    "build_complex_obj_repr",
    "get_input_image",
    "build_deprecation_message",
    "get_device",
    "download_file",
    "reduce",
    "suppress_warnings",
]


def prod(iterable: Iterable) -> Any:
    return _reduce(mul, iterable)


T = TypeVar("T")


def _to_nd_arg(dims: int) -> Callable[[Union[T, Sequence[T]]], Tuple[T, ...]]:
    def to_nd_arg(x: Union[T, Sequence[T]]) -> Tuple[T, ...]:
        if isinstance(x, Sequence):
            if len(x) != dims:
                msg = f"Expected sequence of length {dims}, but got {len(x)}."
                raise RuntimeError(msg)
            return tuple(x)
        else:
            return tuple(itertools.repeat(x, dims))

    return to_nd_arg


to_1d_arg = _to_nd_arg(1)
to_2d_arg = _to_nd_arg(2)
to_3d_arg = _to_nd_arg(3)


def zip_equal(*sequences: Sequence) -> Iterable:
    numel = len(sequences[0])
    if any(len(sequence) != numel for sequence in sequences[1:]):
        raise RuntimeError("All sequences should have the same length")
    return zip(*sequences)


# FIXME: add padding
# FIXME: add sign
#  see https://pyformat.info/#param_align
def build_fmtstr(
    id: Optional[Union[int, str]] = None,
    align: Optional[str] = None,
    field_len: Optional[Union[int, str]] = None,
    precision: Optional[Union[int, str]] = None,
    type: Optional[str] = None,
) -> str:
    fmtstr = r"{"
    if id is not None:
        fmtstr += str(id)
    fmtstr += ":"
    if align is not None:
        fmtstr += align
    if field_len is not None:
        fmtstr += str(field_len)
    if precision is not None:
        fmtstr += "." + str(precision)
    if type is not None:
        fmtstr += type
    fmtstr += r"}"
    return fmtstr


def verify_str_arg(
    arg: Any, param: Optional[str] = None, valid_args: Optional[Sequence[str]] = None
) -> str:
    if not isinstance(arg, str):
        if param is None:
            msg1 = "Expected type str"
        else:
            msg1 = f"Expected type str for parameter {param}"
        msg2 = f", but got type {type(arg)}."
        raise ValueError(msg1 + msg2)

    if valid_args is None:
        return arg

    if arg not in valid_args:
        if param is None:
            msg1 = f"Unknown argument '{arg}'. "
        else:
            msg1 = f"Unknown argument '{arg}' for parameter {param}. "
        msg2 = "Valid arguments are {{{}}}."
        msg2 = msg2.format("'" + "', '".join(valid_args) + "'")
        raise ValueError(msg1 + msg2)

    return arg


def build_complex_obj_repr(
    name: str,
    properties: Optional[Dict[str, Any]] = None,
    named_children: Sequence[Tuple[str, Any]] = (),
    line_length: int = 80,
    num_indent: int = 2,
) -> str:
    def format_properties(properties: Dict[str, Any], sep: str) -> str:
        return sep.join([f"{key}={value}" for key, value in properties.items()])

    def indent(line: str) -> str:
        return " " * num_indent + line

    if properties is None:
        properties = {}

    prefix = f"{name}("
    postfix = ")"

    body = format_properties(properties, ", ")

    body_too_long = (
        len(body) + (num_indent if named_children else len(prefix) + len(postfix))
        > line_length
    )
    multiline_body = len(str(body).splitlines()) > 1

    if body_too_long or multiline_body:
        body = format_properties(properties, ",\n")
    elif not named_children:
        return prefix + body + postfix

    body = [indent(line) for line in body.splitlines()]

    for name, module in named_children:
        lines = str(module).splitlines()
        body.append(indent(f"({name}): {lines[0]}"))
        for line in lines[1:]:
            body.append(indent(line))

    return "\n".join([prefix] + body + [postfix])


def get_input_image(
    starting_point: Union[str, torch.Tensor] = "content",
    content_image: Optional[torch.Tensor] = None,
    style_image: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Generates an input image for NST from the given ``starting_point``.

    Args:
        starting_point: If :class:`~torch.Tensor` returns a copy. If ``"content"`` or
         ``"style"`` returns a copy of ``content_image`` or ``style_image``,
         respectively. If ``"random"`` returns a white noise image with the dimensions
         of ``content_image`` or ``style_image``, respectively. Defaults to
         ``"content"``.
        content_image: Content image. Only required if ``starting_point`` is
            ``"content"`` or ``"random"``.
        style_image: Style image. Only required if ``starting_point`` is
            ``"style"`` or ``"random"``.
    """
    if isinstance(starting_point, torch.Tensor):
        return starting_point.clone()

    starting_point = verify_str_arg(
        starting_point, "starting_point", ("content", "style", "random")
    )

    if starting_point == "content":
        if content_image is not None:
            return content_image.clone()
        raise RuntimeError("starting_point is 'content', but no content image is given")
    elif starting_point == "style":
        if style_image is not None:
            return style_image.clone()
        raise RuntimeError("starting_point is 'style', but no style image is given")
    else:  # starting_point == "random":
        if content_image is not None:
            return torch.rand_like(content_image)
        elif style_image is not None:
            return torch.rand_like(style_image)
        raise RuntimeError("starting_point is 'random', but no image is given")


def build_deprecation_message(
    description: str,
    version: str,
    info: Optional[str] = None,
    url: Optional[str] = None,
) -> str:
    msg = (
        f"{description.strip()} is deprecated since pystiche=={version} and will be "
        "removed in a future release."
    )
    if info is not None:
        msg += f" {info.strip()}"
    if url is not None:
        msg += f" See {url} for further details."
    return msg


def get_device(device: Optional[str] = None) -> torch.device:
    """Selects a device to perform an NST on.

    Args:
        device: If ``str``, returns the corresponding :class:`~torch.device`. If
            ``None`` selects CUDA if available and otherwise CPU. Defaults to ``None``.
    """
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    return torch.device(device)


def download_file(
    url: str,
    file: Optional[str] = None,
    user_agent: Optional[str] = None,
    md5: Optional[str] = None,
) -> str:
    if file is None:
        file = path.basename(url)
    if user_agent is None:
        user_agent = "pystiche"
    else:
        warnings.warn(build_deprecation_message("The parameter user_agent", "0.6.0"))

    request = Request(url, headers={"User-Agent": user_agent})

    try:
        with urlopen(request) as response, open(file, "wb") as fh:
            fh.write(response.read())
    except HTTPError as error:
        msg = f"The server returned {error.code}: {error.reason}."
        raise RuntimeError(msg) from error

    if md5 is not None and not check_md5(file, md5):
        raise RuntimeError(f"The MD5 checksum of {file} mismatches.")

    return file


def reduce(x: torch.Tensor, reduction: str) -> torch.Tensor:
    """Reduces a :class:`~torch.Tensor` as specified.

    Args:
        x: Input tensor.
        reduction: Reduction method to be applied to ``x``. If ``"none"``, no reduction
            will be applied. If ``"sum"`` or ``"mean"``, the :func:`~torch.sum` or
            :func:`~torch.mean` will be applied across all dimensions of ``x``.
    """
    verify_str_arg(reduction, "reduction", ("mean", "sum", "none"))
    if reduction == "mean":
        return torch.mean(x)
    elif reduction == "sum":
        return torch.sum(x)
    else:  # reduction == "none":
        return x


@contextlib.contextmanager
def suppress_warnings(*categories: Type[Warning]) -> Iterator[None]:
    if not categories:
        categories = (UserWarning,)
    old_filters = set(warnings.filters)  # type: ignore[attr-defined]
    for category in categories:
        warnings.filterwarnings("ignore", category=category)
    new_filters = set(warnings.filters) - old_filters  # type: ignore[attr-defined]
    try:
        yield
    finally:
        for filter in new_filters:
            warnings.filters.remove(filter)  # type: ignore[attr-defined]
        warnings._filters_mutated()  # type: ignore[attr-defined]



from typing import Any, Dict

import torch

from . import functional as F
from .op import PixelRegularizationOperator

__all__ = ["TotalVariationOperator", "ValueRangeOperator"]


class TotalVariationOperator(PixelRegularizationOperator):
    r"""The total variation loss is a regularizer used to suppress checkerboard
    artifacts by penalizing the gradient of the image. It is calculated by

    .. math::

        \mean[\limits_{i,j}] \parentheses{\parentheses{x_{i,j+1} - x_{i,j}}^2 + \parentheses{x_{i+1,j} - x_{i,j}}^2}^{\frac{\beta}{2}}

    where :math:`x` denotes the image and :math:`i,j` index a specific pixel.

    .. note::

        Opposed to the paper, the implementation calculates the grand average
        :math:`\mean` opposed to the grand sum :math:`\sum` to account for different
        sized images.

    Args:
        exponent: Parameter :math:`\beta` . A higher value leads to more smoothed
            results. Defaults to ``2.0``.
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Examples:

        >>> op = ops.TotalVariationOperator()
        >>> input = torch.rand(2, 3, 256, 256)
        >>> score = op(input)

    .. seealso::

        The total variation loss was introduced by Mahendran and Vedaldi in
        :cite:`MV2015` .
    """

    def __init__(self, exponent: float = 2.0, score_weight: float = 1.0):
        super().__init__(score_weight=score_weight)
        self.exponent = exponent

    def input_image_to_repr(self, image: torch.Tensor) -> torch.Tensor:
        return image

    def calculate_score(self, input_repr: torch.Tensor) -> torch.Tensor:
        return F.total_variation_loss(input_repr, exponent=self.exponent)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["exponent"] = f"{self.exponent:g}"
        return dct


class ValueRangeOperator(PixelRegularizationOperator):
    def input_image_to_repr(self, image: torch.Tensor) -> torch.Tensor:
        return image

    def calculate_score(self, input_repr: torch.Tensor) -> torch.Tensor:
        return F.value_range_loss(input_repr)



from typing import Dict, Iterable

import torch

from pystiche.ops import ComparisonOperator, Operator

__all__ = ["ImageStorage"]


class ImageStorage:
    def __init__(self, ops: Iterable[Operator]) -> None:
        self.target_guides: Dict[ComparisonOperator, torch.Tensor] = {}
        self.target_images: Dict[ComparisonOperator, torch.Tensor] = {}
        self.input_guides: Dict[Operator, torch.Tensor] = {}
        for op in ops:
            if isinstance(op, ComparisonOperator):
                if op.has_target_guide:
                    self.target_guides[op] = op.target_guide

                if op.has_target_image:
                    self.target_images[op] = op.target_image

            if op.has_input_guide:
                self.input_guides[op] = op.input_guide

    def restore(self) -> None:
        for op, target_guide in self.target_guides.items():
            op.set_target_guide(target_guide, recalc_repr=False)

        for op, target_image in self.target_images.items():
            op.set_target_image(target_image)

        for op, input_guide in self.input_guides.items():
            op.set_input_guide(input_guide)



from os import path
from typing import Any, Optional, Tuple, Union, cast

from PIL import Image

import torch
from torchvision.transforms.functional import to_pil_image as _to_pil_image
from torchvision.transforms.functional import to_tensor as _to_tensor

from .utils import (
    calculate_aspect_ratio,
    edge_to_image_size,
    extract_batch_size,
    force_image,
    force_single_image,
    is_batched_image,
    is_edge_size,
    is_image_size,
    make_batched_image,
    make_single_image,
)

try:
    import matplotlib.pyplot as plt

    def _show_pil_image(image: Image.Image, title: Optional[str] = None) -> None:
        fig, ax = plt.subplots()
        ax.axis("off")
        if title is not None:
            ax.set_title(title)

        cmap = "gray" if image.mode in ("L", "1") else None
        ax.imshow(image, cmap=cmap)
        plt.show()


except ImportError:

    def _show_pil_image(image: Image.Image, title: Optional[str] = None) -> None:
        image.show(title=title)


__all__ = [
    "import_from_pil",
    "export_to_pil",
    "read_image",
    "write_image",
    "show_image",
]

_PIL_RESAMPLE_MAP = {
    "nearest": Image.NEAREST,
    "bilinear": Image.BILINEAR,
    "bicubic": Image.BICUBIC,
}


def import_from_pil(
    image: Image.Image,
    device: Union[torch.device, str] = "cpu",
    make_batched: bool = True,
) -> torch.Tensor:
    if isinstance(device, str):
        device = torch.device(device)
    image = cast(torch.Tensor, _to_tensor(image).to(device))
    if make_batched:
        image = make_batched_image(image)
    return image


@force_image
def export_to_pil(
    image: torch.Tensor, mode: Optional[str] = None
) -> Union[Image.Image, Tuple[Image.Image, ...]]:
    def fn(image: torch.Tensor) -> Image.Image:
        return _to_pil_image(image.detach().cpu().clamp(0.0, 1.0), mode)

    if is_batched_image(image):
        batched_image = image
        batch_size = extract_batch_size(batched_image)
        if batch_size == 1:
            return fn(make_single_image(batched_image))
        else:
            return tuple([fn(single_image) for single_image in batched_image])

    return fn(image)


def _pil_resize(
    image: Image.Image,
    size: Union[int, Tuple[int, int]],
    interpolation_mode: str,
    **kwargs: Any,
) -> Image.Image:
    if is_image_size(size):
        size = cast(Tuple[int, int], size)
        height, width = size
    elif is_edge_size(size):
        size = cast(int, size)
        height, width = edge_to_image_size(
            size, calculate_aspect_ratio((image.height, image.width))
        )
    else:
        msg = (
            f"size can either be an edge size (int) or an image size "
            f"(Tuple[int, int]), but got {type(size)}."
        )
        raise TypeError(msg)

    return image.resize((width, height), resample=_PIL_RESAMPLE_MAP[interpolation_mode])


def read_image(
    file: str,
    device: Union[torch.device, str] = "cpu",
    make_batched: bool = True,
    size: Optional[Union[int, Tuple[int, int]]] = None,
    interpolation_mode: str = "bilinear",
) -> torch.Tensor:
    """Read an image from file with :mod:`PIL.Image` and return it as
    :class:`~torch.Tensor` .

    Args:
        file: Path to image file to be read.
        device: Device that the image is transferred to. Defaults to CPU.
        make_batched: If ``True``, a fake batch dimension is added to the image.
        size: Optional size the image is resized to.
        interpolation_mode: Interpolation mode that is used to perform the optional
            resizing. Valid modes are ``"nearest"``, ``"bilinear"``, and ``"bicubic"``.
            Defaults to ``"bilinear"``.
    """
    if isinstance(device, str):
        device = torch.device(device)

    image = Image.open(path.expanduser(file))

    if size is not None:
        image = _pil_resize(image, size, interpolation_mode)

    return import_from_pil(image, device=device, make_batched=make_batched)


@force_single_image
def write_image(
    image: torch.Tensor, file: str, mode: Optional[str] = None, **save_kwargs: Any
) -> None:
    """Write a :class:`~torch.Tensor` image to a file with :mod:`PIL.Image` .

    Args:
        image: Image to be written.
        file: Path to image file.
        mode: Optional image mode. See the `Pillow documentation
            <https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes>`_
            for details.
        **save_kwargs: Other parameters that are passed to :meth:`PIL.Image.Image.save`
            .
    """
    export_to_pil(image, mode=mode).save(file, **save_kwargs)


def show_image(
    image: Union[torch.Tensor, str],
    title: Optional[str] = None,
    mode: Optional[str] = None,
    size: Optional[Union[int, Tuple[int, int]]] = None,
    interpolation_mode: str = "bilinear",
) -> None:
    """Show an image and optionally read it from file first.

    .. note::

        ``show_image`` uses :func:`matplotlib.pyplot.imshow` as primary means to show
        images. If that is not available the native :meth:`PIL.Image.Image.show` is
        used as a fallback.

    Args:
        image: Image to be shown. If ``str`` this is treated as a path to an image and
            is read by :func:`~pystiche.image.read_image` .
        title: Optional title of the image.
        mode: Optional image mode. See the `Pillow documentation
            <https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes>`_
            for details.
        size: Optional size the image is resized to.
        interpolation_mode: Interpolation mode that is used to perform the optional
            resizing. Valid modes are ``"nearest"``, ``"bilinear"``, and ``"bicubic"``.
            Defaults to ``"bilinear"``.
    """
    if isinstance(image, torch.Tensor):
        image = export_to_pil(image, mode=mode)
    elif isinstance(image, str):
        image = Image.open(path.expanduser(image))
    else:
        msg = f"image can either be torch.Tensor or str, but got {type(image)}."  # type: ignore[unreachable]
        raise TypeError(msg)

    if size is not None:
        image = _pil_resize(image, size, interpolation_mode)

    _show_pil_image(image, title=title)



import itertools
import warnings
from typing import Any, Iterable, Union

import torch

import pystiche
from pystiche.misc import build_deprecation_message

__all__ = ["Transform", "ComposedTransform"]


class Transform(pystiche.Module):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        msg = build_deprecation_message(
            "Using functionality from pystiche.image.transforms",
            "0.7.0",
            info="See https://github.com/pmeier/pystiche/issues/382 for details.",
        )
        warnings.warn(msg, UserWarning)
        super().__init__(*args, **kwargs)

    def __add__(
        self, other: Union["Transform", "ComposedTransform"]
    ) -> "ComposedTransform":
        return compose_transforms(self, other)


class ComposedTransform(Transform):
    def __init__(self, *transforms: Transform):
        super().__init__(indexed_children=transforms)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for module in self.children():
            x = module(x)
        return x

    def __add__(
        self, other: Union["Transform", "ComposedTransform"]
    ) -> "ComposedTransform":
        return compose_transforms(self, other)


def compose_transforms(
    *transforms: Union[Transform, ComposedTransform]
) -> ComposedTransform:
    def unroll(
        transform: Union[Transform, ComposedTransform]
    ) -> Iterable[Union[Transform, ComposedTransform]]:
        if isinstance(transform, ComposedTransform):
            return [
                transform
                for transform in transform.children()
                if isinstance(transform, Transform)
            ]
        elif isinstance(transform, Transform):
            return (transform,)
        else:
            msg = (  # type: ignore[unreachable]
                f"transforms can either be pystiche.image.transforms.Transform or "
                f"pystiche.image.transforms.ComposedTransform, but got "
                f"{type(transform)}."
            )
            raise TypeError(msg)

    return ComposedTransform(*itertools.chain(*map(unroll, transforms)))
